/* as21.c  a simple assembler for s21

format  (case insensitive)

	.symbol                   define symbol
	symbol value
	 ...
	.code ads                 code segment at ads
	[:label] opcode operand
	 ...
	.data ads				  data segment at ads
	sym sym ...               define words
	.end

operand
L-format: r1 ads
D-format: r1 r2 disp
X-format: r1 r2 r3 xop

addressing mode
absolute: r1 ads
indirect:  r1 @disp r2
index: r1 +r2 r3
register:  r1 r2 r3
immediate: r1 r2 #sym

operand = sym @sym +sym #sym

modify from cpu3 assembler A3  version 1.0  7th September 1999
fix bug : undef op, add listing 			2 Oct 1999
add extended instructions  					28 November 2001
modify from as2								16 Jan 2007

Prabhas Chongstitvatana
*/

#include "as21.h"

char 	lbuf[MXBUF];				// copy of input
char 	ibuf[MXBUF], *cp = NULL;   	// input buffer, ch ptr
char	ob[MXBUF];					// output buffer
char	cbuf[MXBUF];				// output buffer for listing

static 	char sep[] = " \t\n";      	// separator char
char    *w;							// current input word
int 	lineno,loc;		   			// line num, current ads
FILE 	*fi, *fo, *fl;
token	mem[MXMEM];					// store tokens for pass2
int 	tp = 0;						// token index
int 	mark;						// index to current op token
int		pass;

// initial symbol in the symbol table

typedef struct { char name[8]; int value; int arg; } symtype;

symtype	initsym[] = {
	{"NOP",NOP,NA},{"LD",LD,MEM},{"ST",ST,MEM},{"MV",MV,MOV},
	{"JMP",JMP,SP},{"JAL",JAL,ABS},{"RET",RET,SP},
	{"JT",JT,ABS},{"JF",JF,ABS},
	{"ADD",ADD,RRI},{"SUB",SUB,RRI},{"MUL",MUL,RRI},{"DIV",DIV,RRI},
	{"AND",AND,RRI},{"OR",OR,RRI},{"XOR",XOR,RRI},
	{"EQ",EQ,RRI},{"NE",NE,RRI},{"LT",LT,RRI}, {"LE",LE,RRI},
	{"GT",GT,RRI},{"GE",GE,RRI},
	{"SHL",SHL,RRI},{"SHR",SHR,RRI},{"TRAP",TRAP,SP},
	{"XL",XL,ABS}, {"XD",XD,RI}, {"XX",XX,RR},
	{"R0",0,0},{"R1",1,0},{"R2",2,0},{"R3",3,0},{"R4",4,0},
	{"R5",5,0},{"R6",6,0},{"R7",7,0},{"R8",8,0},{"R9",9,0},
	{"R10",10,0},{"R11",11,0},{"R12",12,0},{"R13",13,0},{"R14",14,0},
	{"R15",15,0},{"R16",16,0},{"R17",17,0},{"R18",18,0},{"R19",19,0},
	{"R20",20,0},{"R21",21,0},{"R22",22,0},{"R23",23,0},{"R24",24,0},
	{"R25",25,0},{"R26",26,0},{"R27",27,0},{"R28",28,0},{"R29",29,0},
	{"R30",30,0},{"R31",31,0},
//	{"ALWAYS",0},{"EQ",1}, {"NEQ",2},{"LT",3},{"LE",4},{"GE",5},{"GT",6},
	{"",0,0}
};

void error(char *s){
	if(pass == 1)
		printf("line %d error: %s symbol %s\n%s",lineno,s,w,lbuf);
	else
		printf("line %d error: %s\n",mem[tp].line,s);
	exit(0);
}
// get one token from input
char *tok(void){
	if( cp != NULL ) cp = strtok(NULL,sep);
	while ( cp == NULL || eqs(cp,";;") ){
		if( fgets(ibuf,MXBUF,fi) == NULL ) return NULL;
		strcpy(lbuf,ibuf);
		lineno++;
		cp = strtok(ibuf, sep);
	}
	cp = strupr(cp);   // convert to uppercase
	return cp;
}
// store token
void store(char type, char mode, int ref){
	if( tp >= MXMEM ) error("out of memory");
	mem[tp].type = type;
	mem[tp].mode = mode;
	mem[tp].ref = ref;
	mem[tp].line = lineno;
	tp++;
}

// check number
int isnumber(char *s){
	int i = 0;
	if(s[0] == '-') i = 1;
	while( isdigit(s[i]) ) i++;
	if( s[i] == 0 ) return 1;
	return 0;
}

int getNum(void){
	w = tok();
	if(isnumber(w))	return atoi(w);
	error("expect number");
	return 0;
}

// set mode of op token
void setop(char mode){
	mem[mark].mode = mode;
}

int valueof(char *s, int *type){
	*type = NUM;
	if( isnumber(s) ) return atoi(s);
	*type = SYM;
	return putsym(s);
}

// read reg, mode RR
void readR(void){
	int v,type;
	w = tok();
	if(w[0]==':' || w[0]=='@' || w[0]=='+' || w[0]=='#')
		error("expect register or ads argument");
	v = valueof(w,&type);
	store(type,RR,v);
}

void storew(char *s, int mode){
	int v,type;
	v = valueof(s,&type);
	store(type,mode,v);
	setop(mode);
}
// read mem, set mode ABS IND INX
void readM(void){
	w = tok();
	switch( w[0] ) {
	case '@': storew(w+1,IND); readR(); break;
	case '+': storew(w+1,INX); readR(); break;
	default: storew(w,ABS); break;
	}
}
// read reg or imm set mode
void readRI(int imm, int r){
	w = tok();
	if( w[0]==':' || w[0]=='@' || w[0]=='+' )
		error("expect register or immediate argument");
	if( w[0] == '#') storew(w+1,imm);
	else storew(w,r);
}
// read and store ex. instruction in mode
void readOP(int mode){
	w = tok();
	storew(w,mode);
}
// read argument according to OP
void readcode(int op){
	int v, type;
	mark = tp;
	store(OP,NA,op);
	switch( op ) {
	case LD: readR(); readM(); break;
	case ST: readM(); readR(); break;
	case JMP:			// one arg
	case RET:
	case TRAP: readR(); setop(SP); break;
	case JAL:			// two arg
	case JT:
	case JF: readR(); readR(); setop(ABS); break;
	case ADD:			// three arg
	case SUB:
	case MUL:
	case DIV:
	case AND:
	case OR:
	case XOR:
	case EQ:
	case NE:
	case LT:
	case LE:
	case GT:
	case GE:
	case SHL:
	case SHR: readR(); readR(); readRI(RI,RR); break;
	case MV: readR(); readRI(IMM,REG); break;
	// user defined instruction
	case XL: readOP(ABS); readR(); readR(); break;
	case XD: readOP(RI); readR(); readR(); readR(); break;
	case XX: readOP(RR); readR(); readR(); readR(); break;
	default: error("undefine op");
	}
}
// check arguments of current op
// type is the specified type, mode is the actual arg
void chkarg(int type){
	int mode, f;
	mode = mem[mark].mode;
	f = 0;
	switch(type){
	case MEM: f = (mode==ABS) || (mode==IND) || (mode==INX); break;
	case RRI: f = (mode==RR)  || (mode==RI); break;
	case MOV: f = (mode==IMM) || (mode==REG);  break;
	case ABS: f = mode == ABS; break;
	case RI:  f = mode == RI; break;
	case RR:  f = mode == RR; break;
	case SP:  f = mode == SP; break;
	}
	if( !f ) error("incorrect argument type");
}

void pass1(void){
	int idx, state;
	pass = 1;
	state = 'S';
	w = tok();
	while( !eqs(w,".END") ) {
		if( eqs(w,".SYMBOL")){
			state = 'S';
		}else if(eqs(w,".DATA")){
			loc = getNum();
			store(DOTD,NA,loc);
			state = 'D';
		}else if(eqs(w,".CODE")){
			loc = getNum();
			store(DOTC,NA,loc);
			state = 'C';
		}else
		switch(state){
		case 'S':
			idx = putsym(w);
			if(getValue(idx) != UNDEF)
				error("duplicate symbol");
			setsym(idx,getNum(),SYM,NA);
			break;
		case 'D':
			if( isnumber(w) )
				store(NUM,NA,atoi(w));
			else
				store(SYM,NA,putsym(w));
			loc++;
			break;
		case 'C':
			if( w[0] == ':' ){	// insert symbol
				idx = putsym(w+1);
				if(getValue(idx) != UNDEF)
					error("duplicate label");
				setsym(idx,loc,SYM,NA);
			}else{
				idx = searchsym(w);
				if(idx < 0 || getType(idx) != OP)
					error("undefined op");
				readcode(getValue(idx));
				chkarg(getArg(idx));
				loc++;
			}
		} // switch
		w = tok();
	}  // while
	store(DOTE,NA,loc);
}

int isdot(char type){
	return type == DOTC || type == DOTD || type == DOTE;
}

/*  opcode encoding
LD   ld a 1, ld d 2, ld x 31 16
ST 	 st a 3, st d 4, st x 31 17
JMP	 jmp 6
JAL  jal 7
JT   jt 8
JF   jf 9
RET	 ret 31 18
mv   mv i 5, mv r 31 15
ADD	9  add i 10, add r 31 0
SUB	10 sub i 11, sub r 31 1
MUL	11 mul i 12, mul r 31 2
DIV	12 div i 13, div r 31 3
AND	13 and i 14, and r 31 4
OR  14 or  i 15, or  r 31 5
XOR 15 xor i 16, xor r 31 6
EQ  16 ...
NE  17
LT  18
LE  19
GT  20
GE  21
SHL 22
SHR 23
TRAP 24	trap 31 19
*/

int rdtokval(void){
	int v;
	v = UNDEF;
	switch( mem[tp].type ){
	case NUM: v = mem[tp].ref; break;
	case SYM: v = getValue(mem[tp].ref);
	}
	if(v == UNDEF) error("undefined symbol");
	tp++;
	return v;
}

void readarg(int mode, int *a1, int *a2, int *a3){
	*a1 = rdtokval();	// read one arg
	switch( mode ){
	case SP: break;		// one arg
	case ABS:			// two arg
	case IMM:
	case REG: *a2 = rdtokval(); break;
	case INX:			// three arg
	case IND:
	case RI:
	case RR: *a2 = rdtokval(); *a3 = rdtokval(); break;
	default: error("unknow addressing mode");
	}
}

PRIVATE void prL(int op, int a1, int a2){
	sprintf(ob,"L %d %d %d",op,a1,a2);
}
PRIVATE void prD(int op, int a1, int a2, int a3){
	sprintf(ob,"D %d %d %d %d",op,a1,a2,a3);
}
PRIVATE void prX(int xop, int a1, int a2, int a3){
	sprintf(ob,"X 31 %d %d %d %d",a1,a2,a3,xop);
}

void gencode(void){
	int op, op2, mode, a1, a2, a3;
	mode = mem[tp].mode;
	op = mem[tp].ref;
	tp++;
	if( op==XL || op==XD || op==XX ) op2 = rdtokval();
	readarg(mode,&a1,&a2,&a3);
	switch( op ){
	case LD:
		switch( mode ){
		case ABS: prL(1,a1,a2); break;
		case IND: prD(2,a1,a3,a2); break;
		case INX: prX(16,a1,a2,a3); break;
		}
		break;
	case ST:
		switch( mode ){
		case ABS: prL(3,a2,a1); break;
		case IND: prD(4,a3,a2,a1); break;
		case INX: prX(17,a3,a1,a2); break;
		}
		break;
	case JMP: prL(6,0,a1); break;
	case JAL: prL(7,a1,a2); break;
	case JT:  prL(8,a1,a2); break;
	case JF:  prL(9,a1,a2); break;
	case RET: prX(18,a1,0,0); break;
	case ADD:
	case SUB:
	case MUL:
	case DIV:
	case AND:
	case OR:
	case XOR:
	case EQ:
	case NE:
	case LT:
	case LE:
	case GT:
	case GE:
	case SHL:
	case SHR:
		switch( mode ){
		case RI: prD(op+1,a1,a2,a3); break;
		case RR: prX(op-9,a1,a2,a3); break;
		}
		break;
	case MV:
		switch( mode ){
		case IMM: prL(5,a1,a2); break;
		case REG: prX(15,a1,a2,0); break;
		}
		break;
	case TRAP: prX(19,a1,0,0); break;
	case XL: prL(op2,a1,a2); break;
	case XD: prD(op2,a1,a2,a3); break;
	case XX: prX(op2,a1,a2,a3); break;
	default: error("undefine op");
	}
}

void outobj(void){
	fprintf(fo,"%s\n",ob);
}

void outlst(void){
	int i, line;
	line = mem[tp-1].line;
	// get source until before current line
	while(lineno < (line-1)){
		fgets(lbuf,MXBUF,fi);
		fprintf(fl,"                        %s",lbuf);
		lineno++;
	}
	fgets(lbuf,MXBUF,fi);	// current line
	lineno++;
	sprintf(cbuf,"%4d %s",loc,ob);
	for(i=strlen(cbuf); i<24; i++)	// pad blank to 24
		cbuf[i] = 32;
	strcpy(cbuf+24,lbuf);
	fprintf(fl,"%s",cbuf);
}

void pass2(void){
	pass = 2;
	tp = 0;
	lineno = 0;
	rewind(fi);
	while( mem[tp].type != DOTE){
		switch(mem[tp].type){
		case DOTC:
			loc = mem[tp].ref;
			sprintf(ob,"a %d",loc);
			outobj();
			tp++;
			while(!isdot(mem[tp].type)){
				gencode();
				outobj();
				outlst();
				loc++;
			}
			break;
		case DOTD:
			loc = mem[tp].ref;
			sprintf(ob,"a %d",loc);
			outobj();
			tp++;
			while(!isdot(mem[tp].type)){
				sprintf(ob,"w %d",rdtokval());
				outobj();
				loc++;
			}
		}
	}
	sprintf(ob,"e");
	outobj();
	// list rest of the source
	while(fgets(lbuf,MXBUF,fi) != NULL)
		fprintf(fl,"                        %s",lbuf);
}

// put reserved words into symbol table
void initsymbol(void){
	int i;
	for(i=0; initsym[i].name[0] != 0; i++)
		setsym(putsym(initsym[i].name),
			initsym[i].value,OP,initsym[i].arg);
}

// make an obj file name from source
void makename( char *source, char *obj, char *lis ){
	int n;
	n = strcspn(source,".");
	strncpy(obj,source,n);
	strcpy(obj+n,".obj");
	strncpy(lis,source,n);
	strcpy(lis+n,".lis");
}

void main(int argc, char *argv[]){
	char fout[80], flis[80];

	if( argc < 2 ) {
		printf("usage : as21 inputfile\n");
		exit(0);
	}
	fi = fopen(argv[1],"r");
	if( fi == NULL ){
		printf("input file not found\n");
		exit(0);
	}
	makename(argv[1], fout, flis);
	initsymbol();
	lineno = 0;
	loc = 0;
	pass1();
	fl = fopen(flis,"w");
	fo = fopen(fout,"w");
	pass2();
	fclose(fi);
	fclose(fo);
	fclose(fl);
}

/*----------- for debugging


void testread(void){
	int cnt = 100;
	w = tok();
	while ( !eqs(w,".END") && (w != NULL) ){
		printf("%s ",w);
		w = tok();
		if(cnt-- < 0) break;
	}
}

char prmode( int mode){
	switch( mode ){
	case ABS: return 'A';
	case DISP:return '@';
	case INX:return '+';
	case IMM:return '%';
	case RI:return '#';
	case RR:return 'R';
	case SP:return 'S';
	case NA:return '?';
	case EL: return 'L';
	case ED: return 'D';
	case EX: return 'X';
	}
	return 0;
}

void prtoken(int type, int mode, int ref){
	switch( type ){
	case SYM: printf("sym %c %d\n",prmode(mode),ref); break;
	case NUM: printf("num %c %d\n",prmode(mode),ref); break;
	case OP: printf("op %c %d\n",prmode(mode),ref); break;
	case DOTA: printf(".a %d\n",ref); break;
	case DOTC: printf(".c %d\n",ref); break;
	case DOTW: printf(".w %d\n",ref); break;
	case DOTE: printf(".e\n"); break;
	}
}

void dumptoken(){
	int i;
	for(i=0;i<tp;i++)
		prtoken(mem[i].type,mem[i].mode,mem[i].ref);
}

--------------- */
